# Copyright (c) 2013, Charles O. Goddard and Alethea Butler.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from collections import namedtuple
from io import BytesIO

from .bitstream import BitStream


class Scope(dict):
    '''
    A lexical scope.

    Contains a mapping of names to objects and a parent scope.
    '''
    def __init__(self, *args, parent=None, **kwargs):
        self.parent = parent
        super(Scope, self).__init__(*args, **kwargs)

    def __getitem__(self, key):
        try:
            return super(Scope, self).__getitem__(key)
        except KeyError:
            if self.parent:
                return self.parent[key]
            raise

    def __contains__(self, key):
        if super(Scope, self).__contains__(key):
            return True
        elif self.parent:
            return key in self.parent
        return False


class Type:
    def __init__(self, name):
        self.name = name

    def prepare(self, dict):
        pass

    def deserialize(self, bs, parent_scope):
        raise NotImplementedError()

    def serialize(self, bs, value, parent_scope):
        raise NotImplementedError()

    def frombytes(self, data):
        bio = BytesIO(data)
        return self.deserialize(BitStream(bio))

    def tobytes(self, value):
        bio = BytesIO()
        self.serialize(BitStream(bio), value)
        return bio.getvalue()

    def __str__(self):
        return '{0}: <{1}>'.format(self.name, type(self).__name__)


class Structure(Type):
    def __init__(self, name, fields, *args, **kwargs):
        self.tuple = namedtuple(name, ' '.join(f.name for f in fields))
        self.fields = fields
        super(Structure, self).__init__(name, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.tuple(*args, **kwargs)

    def deserialize(self, bs, parent_scope=None):
        scope = Scope(parent=parent_scope)
        for f in self.fields:
            scope[f.name] = f.deserialize(bs, scope)
        return self.tuple(**scope)

    def serialize(self, bs, value, parent_scope=None):
        scope = Scope(value.__dict__, parent=parent_scope)
        for f in self.fields:
            f.serialize(bs, scope[f.name], scope)

    def __str__(self):
        return '{0}: | {1} |'.format(
            self.name, ' | '.join([str(f) for f in self.fields]))


class BranchPoint(Type):
    def __init__(self, name, cases,  *args, **kwargs):
        self.cases = cases
        super(BranchPoint, self).__init__(name, *args, **kwargs)

    def deserialize(self, bs, parent_scope=None):
        scope = Scope(parent=parent_scope)
        for predicate, result in self.cases.items():
            if predicate(scope):
                return result.deserialize(bs, scope)

    def serialize(self, bs, value, parent_scope=None):
        scope = Scope(value.__dict__, parent=parent_scope)
        for predicate, result in self.cases.items():
            if predicate(scope):
                result.serialize(bs, value, scope)


class Int(Type):
    def __init__(self, name, bits=32, signed=True, *args, **kwargs):
        self.bits = bits
        self.signed = signed
        super(Int, self).__init__(name, *args, **kwargs)

    def deserialize(self, bs, parent_scope):
        return bs.read_int(self.bits, self.signed)

    def serialize(self, bs, value, parent_scope):
        bs.write_int(value, self.bits, self.signed)

    def __str__(self):
        if self.signed:
            signed = 'Signed'
        else:
            signed = 'Unsigned'
        return '{0}: <{1}-bit {2} {3}>'.format(self.name, self.bits,
                                               signed, type(self).__name__)


def make_enum(name, type_, **kwargs):
    python2raw = kwargs
    raw2python = {v: k for k, v in kwargs.items()}

    class Enum(type_):
        def deserialize(self, bs, parent_scope):
            return raw2python[super(Enum, self).deserialize(bs, parent_scope)]

        def serialize(self, bs, value, parent_scope):
            raw = python2raw[value]
            super(Enum, self).serialize(bs, raw, parent_scope)

    Enum.__name__ = name
    return Enum


class Size(Int):
    def __init__(self, name, fieldname, signed=False, *args, **kwargs):
        self.fieldname = fieldname
        super(Size, self).__init__(name, signed=signed, *args, **kwargs)

    def serialize(self, bs, value, parent_scope):
        super(Size, self).serialize(
            bs, len(parent_scope[self.fieldname]), parent_scope)


class Bytes(Type):
    def __init__(self, name, size, padding=0, *args, **kwargs):
        self.size = size
        self.padding = padding
        super(Bytes, self).__init__(name, *args, **kwargs)

    def deserialize(self, bs, parent_scope):
        res = bytearray()
        if self.size is None:
            while True:
                try:
                    c = bs.read_int(bits=8, signed=False)
                    res.append(c)
                except EOFError:
                    break
        else:
            size = self.size
            if isinstance(self.size, Size):
                size = parent_scope[size.name]
            elif isinstance(self.size, str):
                size = parent_scope[size]

            for _ in range(size):
                c = bs.read_int(bits=8, signed=False)
                res.append(c)
        return bytes(res)

    def serialize(self, bs, value, parent_scope):
        if self.size is None or isinstance(self.size, Size):
            for b in value:
                bs.write_int(b, bits=8, signed=False)
        else:
            if len(value) > self.size:
                raise RuntimeError('Would truncate')

            for b in value:
                bs.write_int(b, bits=8, signed=False)
            for _ in range(self.size - len(value)):
                bs.write_int(self.padding, bits=8, signed=False)

    def __str__(self):
        name = type(self).__name__
        if self.size is not None:
            if self.padding != 0:
                return '{0}: <{1} {2} ({3}-padded)>'.format(
                    self.name, self.size, name, self.padding)
            return '{0}: <{1} {2}>'.format(self.name, self.size, name)
        return '{0}: <{1}>'.format(self.name, name)


class CString(Type):
    def deserialize(self, bs, parent_scope):
        res = []
        c = 0
        while True:
            c = bs.read_int(bits=8, signed=False)
            if c == 0:
                break
            res.append(chr(c))
        return ''.join(res)

    def serialize(self, bs, value, parent_scope):
        data = value.encode('latin-1')
        for c in data:
            bs.write_int(c, bits=8, signed=False)
        bs.write_int(0, bits=8, signed=False)
